<?php
/**
 * Extension to highlight the new "greyd-theme".
 * 
 * @since 2.0
 */
if( ! defined( "ABSPATH" ) ) exit;


new Greyd_Theme_Notice();
class Greyd_Theme_Notice {

	/**
	 * Error message.
	 *
	 * @var string
	 */
	public $error = null;

	/**
	 * Constructor
	 */
	public function __construct() {

		if ( ! is_admin() || ! current_user_can( "manage_options" ) ) return false;

		add_action( "admin_init", array($this, "handle_actions") );

		add_action( "admin_notices", array($this, "display_notices"), 99 );
	}

	/**
	 * Handle notice actions
	 */
	public function handle_actions() {

		/**
		 * Save user meta to ignore notice.
		 */
		if ( isset($_GET["greyd_theme_available_ignore"]) && $_GET["greyd_theme_available_ignore"] === "true" ) {
			global $current_user;
			$user_id = $current_user->ID;
			add_user_meta($user_id, "greyd_theme_available_ignore", "true", true);
		}

		if ( isset($_GET["install_greyd_theme"]) ) {

			// install the theme
			if ( $_GET["install_greyd_theme"] === "install" ) {

				$result = $this->install_greyd_theme();

				if ( $result ) {
					// redirect to themes page with activate notice
					wp_safe_redirect( admin_url( "themes.php?install_greyd_theme=installed" ) );
					exit;
				}
			}

			// activate the theme
			if ( $_GET["install_greyd_theme"] === "activate" ) {

				$result = $this->activate_greyd_theme();

				if ( $result ) {
					// redirect to installer
					wp_safe_redirect( admin_url( "admin.php?page=install-greyd" ) );
					exit;
				}
				exit;
			}
		}
	}

	/**
	 * Display admin notices.
	 */
	public function display_notices() {

		// errors
		if ( $this->error ) {
			?>
			<div class="notice notice-error">
				<p><?php echo sprintf( __( 'Beim Installieren des Greyd Theme ist ein Fehler aufgetreten: %s', 'core' ), $this->error ); ?></p>
			</div>
			<?php
			return;
		}
		
		// only display on "themes", "plugins" and "dashboard" pages
		// ...and on "greyd".* pages
		$screen = get_current_screen();
		if (
			! in_array( $screen->id, array( "dashboard", "themes", "plugins" ) )
			&& strpos( $screen->id, "greyd" ) === false
		) {
			return;
		}

		if ( ! isset($_GET["install_greyd_theme"]) ) {
			$this->display_install_notice();
		}
		else if ( $_GET["install_greyd_theme"] === "installed" ) {
			$this->display_activate_notice();
		}

	}

	/**
	 * Display install notice.
	 */
	public function display_install_notice() {

		// check if theme is already installed
		$theme = wp_get_theme( "greyd-theme" );
		if ( $theme->exists() ) return;

		// user dismissed notice
		global $current_user;
		$user_id = $current_user->ID;
		if ( get_user_meta( $user_id, "greyd_theme_available_ignore", true ) ) return;
		
		$this->render_notice_stylesheet();
		?>
		<div id="greyd-notice" class="notice">
			<div>
				<img src="<?php echo get_template_directory_uri() . '/assets/icon/greyd-logo-dark.svg'; ?>" alt="Greyd Logo">
			</div>
			<div>
				<h4><?php echo sprintf( __( 'Hallo %s', 'core' ), self::get_user_welcome_name() ); ?> <img draggable="false" role="img" class="emoji" alt="✌" src="https://s.w.org/images/core/emoji/14.0.0/svg/270c.svg"></h4>
				<p><?php echo __("Schon gewusst? Das neue Greyd.Suite Full Site Editing Theme ist da. Stelle jetzt um und profitiere von zahlreichen neuen Funktionen!", "core"); ?></p>
				<div class="notice-links">
					<a class="button button-greyd" href="<?php echo add_query_arg( "install_greyd_theme", "install" ); ?>"><?php echo __("Theme installieren", "core"); ?></a>
					<a href="<?php echo add_query_arg( "greyd_theme_available_ignore", "true" ); ?>"><?php echo __("diese Meldung verbergen", "core"); ?></a>
				</div>
			</div>
		</div>
		<?php
	}

	/**
	 * Display activate notice.
	 */
	public function display_activate_notice() {

		$this->render_notice_stylesheet();
		?>
		<div id="greyd-notice" class="notice">
			<div>
				<img src="<?php echo get_template_directory_uri() . '/assets/icon/greyd-logo-dark.svg'; ?>" alt="Greyd Logo">
			</div>
			<div>
			<p><?php echo __("Das Greyd Theme wurde erfolgreich installiert. Möchtest du es jetzt aktivieren? Dadurch wird sich deine Website optisch verändern, aber es werden keine Inhalte verändert. Du kannst also jederzeit einfach zurück zum alten Theme wechseln und deine Website sieht wieder so aus wie zuvor.", "core"); ?></p>
				<div class="notice-links">
					<a class="button button-greyd" href="<?php echo add_query_arg( "install_greyd_theme", "activate" ); ?>"><?php echo __("Theme aktivieren", "core"); ?></a>
					<a href="<?php echo remove_query_arg( "install_greyd_theme", add_query_arg( "greyd_theme_available_ignore", "true" ) ); ?>"><?php echo __("diese Meldung verbergen", "core"); ?></a>
				</div>
			</div>
		</div>
		<?php
	}

	/**
	 * =================================================================
	 *                          Utils
	 * =================================================================
	 */

	/**
	 * Install greyd-theme.
	 * 
	 * @return bool
	 */
	public function install_greyd_theme() {

		// check if theme is already installed
		$theme = wp_get_theme( "greyd-theme" );
		if ( ! $theme->exists() ) {
			include_once ABSPATH . 'wp-admin/includes/class-wp-upgrader.php';
			wp_cache_flush();
			
			$upgrader  = new \Theme_Upgrader( new \WP_Ajax_Upgrader_Skin() );
			$installed = $upgrader->install( 'https://update.greyd.io/public/themes/greyd-theme/greyd-theme.zip' );

			if ( is_wp_error( $installed ) ) {
				$this->error = $installed->get_error_message();
			}
		}

		return true;
	}

	/**
	 * Activate greyd-theme.
	 */
	public function activate_greyd_theme() {

		// activate theme
		switch_theme( "greyd-theme" );

		// // re-activate greyd-hub plugin
		// $activated = activate_plugin(
		// 	'greyd_tp_management/init.php',
		// 	'', // redirect
		// 	false, // network_wide
		// 	true // silent
		// );

		// if ( is_wp_error( $activated ) ) {
		// 	$this->error = $activated->get_error_message();
		// 	return false;
		// }

		return true;
	}

	/**
	 * Get name of the current user to welcome with.
	 *
	 * @return string
	 */
	public static function get_user_welcome_name() {

		$user = wp_get_current_user();

		if ( ! empty( $user->first_name ) ) {
			return $user->first_name;
		}

		if ( ! empty( $user->display_name ) ) {
			return $user->display_name;
		}

		return $user->user_nicename;
	}

	/**
	 * Render notice stylesheet.
	 */
	public function render_notice_stylesheet() {
		?>
		<style>
			#greyd-notice {
				--gap: calc(max(1em, 3vh) * 0.66);
				display: flex;
				flex-direction: row;
				padding: var(--gap);
				margin: var(--gap) 0;
				gap: var(--gap);
				border: 1px solid #1e1e1e;
				border-left-width: 10px;
				border-radius: 3px;
				background-color: #ffffff;
				color: #1e1e1e;
				animation: noticePulse 1.2s ease .4s;
			}

			@keyframes noticePulse {
				0% {
					box-shadow: 0 0 0 -4px currentColor;
				}
				100% {
					box-shadow: 0 0 0 6px transparent;
				}
			}

			#greyd-notice :is(h1, h2, h3, h4, h5, h6, p) {
				margin-top: 0;
			}

			#greyd-notice .notice-links {
				display: flex;
				align-items: center;
				gap: var(--gap);
			}

			#greyd-notice a:not(.button) {
				color: #1e1e1e;
			}

			#greyd-notice a:not(.button):hover {
				color: #868A8F;
			}

			#greyd-notice .emoji {
				transform: scale(1.2);
			}
		</style>
		<?php
	}
}